import os
import argparse
from collections import defaultdict
from multiprocessing import Pool

import nltk
import json
import numpy as np
import tensorflow as tf
from bs4 import BeautifulSoup, Comment, NavigableString

import time

# everything that is a child of one of these tags is considered boilerplate and hence ignored by the classifier
TAGS_TO_IGNORE = {'head', 'iframe', 'script', 'meta', 'link', 'style', 'input', 'checkbox', 'button', 'noscript'}

#copied from boilernet.net.preprocess
def get_leaves(node, tag_list=[]):
    """Return all leaves (NavigableStrings) in a BS4 tree."""
    tag_list_new = tag_list + [node.name]
    #if node.has_attr('__boilernet_label'):
    #    label = int(node['__boilernet_label'])

    result = []
    for c in node.children:
        if isinstance(c, NavigableString):
            # might be just whitespace
            if c.string is not None and c.string.strip():
                result.append((c, tag_list_new))
        elif c.name not in TAGS_TO_IGNORE:
            result.extend(get_leaves(c, tag_list_new))
    return result

#copied from boilernet.net.preprocess
def get_leaf_representation(node, tag_list):
    """Return dicts of words and HTML tags that representat a leaf."""
    tags_dict = defaultdict(int)
    for tag in tag_list:
        tags_dict[tag] += 1
    words_dict = defaultdict(int)
    for word in nltk.word_tokenize(node.string):
        words_dict[word.lower()] += 1
    return dict(words_dict), dict(tags_dict)

def get_representation(in_file):
    document = []
    leaves = []
    with open(in_file, 'rb') as hfile:
        doc = BeautifulSoup(hfile, features='html5lib')
        for leaf, tag_list in get_leaves(doc.find_all('html')[0]):
            leaves.append(leaf)
            leaf_representation = get_leaf_representation(leaf, tag_list)
            document.append(leaf_representation)

    # for pickling in multiprocessing
    leaves = [str(leaf) for leaf in leaves]
    return document, leaves

# adapted from boilernet.net.preprocess.get_leaves
def mark_leaves(node, leaves, labels):
    """Mark the leaves with score > 0.5 with the '__boilernet_label'."""
    assert len(leaves) == len(labels)

    for c in node.children:
        # redundant check
        if len(leaves) == 0 and len(labels) == 0:
            return [], []
        if isinstance(c, NavigableString):
            # might be just whitespace
            if c.string is not None and c.string.strip():
                # double check if text correct, then decide based on score
                if c == leaves[0]:
                    if labels[0] > 0.5:
                        node.attrs['__boilernet_label'] = None
                leaves.pop(0)
                labels = labels[1:]
        elif c.name not in TAGS_TO_IGNORE:
            leaves, labels = mark_leaves(c, leaves, labels)
    return leaves, labels

#TODO remove logging of run times
def main():
    ap = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    ap.add_argument('IN_FILES', nargs='+', help='Folder for all processed data files')

    ap.add_argument('--model', default=None, help='Path to model checkpoint')
    ap.add_argument('--tags', default=None, help='Processed JSON with HTML tags')
    ap.add_argument('--words', default=None, help='Processed JSON with all words of the training vocabulary')
    ap.add_argument('--num_workers', default=4, help='Number of parallel processes in file parsing')
    ap.add_argument('--batch_size', default=64, help='Batch size for model inference')
    ap.add_argument('--extract', action='store_true', help='Store annotated html files')
    ap.add_argument('--save', action='store_true', help='Store annotated html files')

    args = ap.parse_args()
    #for i in range(180):
    #    args.IN_FILES.append('model/googletrends2017/gt_' + str(i+1) + '.html')

    start = time.time()

    print('model load')
    t1 = time.time()
    #TODO handling of args.model==None, load default model
    model = tf.keras.models.load_model(args.model)
    print(time.time()-t1)

    print('tag load')
    t1 = time.time()
    #TODO handling of args.tags==None, load default map
    with open(args.tags, 'r', encoding='utf-8') as fp:
        tag_map = json.load(fp)
    print(time.time()-t1)

    print('word load')
    t1 = time.time()
    #TODO handling of args.tags==None, load default map
    with open(args.words, 'r', encoding='utf-8') as fp:
        word_map = json.load(fp)
    print(time.time()-t1)

    # list of lists of all texts (for each document) marked by the model
    clean_docs = []

    # only keep batch_size many inputs in memory
    for i in range(len(args.IN_FILES) // args.batch_size + 1):
        if i < len(args.IN_FILES) // args.batch_size:
            files = args.IN_FILES[i * args.batch_size : (i + 1) * args.batch_size]
        else:
            files = args.IN_FILES[i * args.batch_size :]


        print('get representation')
        t1 = time.time()
        with Pool(args.num_workers) as p:
            all_results = p.map(get_representation, files)
        all_documents = [item[0] for item in all_results]
        all_leaves = [item[1] for item in all_results]
        print(time.time()-t1)

        print("build tf vector")
        t1 = time.time()
        all_features = []
        for doc in all_documents:
            doc_features = []
            for words_dict, tags_dict in doc:
                vocab_vec = np.zeros(len(word_map), dtype='int32')
                for word, num in words_dict.items():
                    # if the item is not in the map, use 0 (OOV word)
                    vocab_vec[word_map.get(word, 0)] = num

                tags_vec = np.zeros(len(tag_map), dtype='int32')
                for tag, num in tags_dict.items():
                    # if the tag is not in the map, use 0 (OOV tag)
                    tags_vec[tag_map.get(tag, 0)] = num

                feature_vector = np.concatenate([vocab_vec, tags_vec])
                doc_features.append(feature_vector)

            # build tensor format for model
            doc_features = np.array(doc_features, dtype='float32', ndmin=2)
            doc_features = tf.stack(doc_features)
            all_features.append(doc_features)
        print(time.time()-t1)

        labels = []

        print('Create ds')
        t1 = time.time()
        # create Dataset to format the batch 
        ds = tf.data.Dataset.from_generator(lambda: all_features, tf.float32, output_shapes=[None, 1052])
        padding_ds = tf.data.Dataset.padded_batch(ds, 64, [None, len(word_map.items()) + len(tag_map.items())])
        print(time.time()-t1)
        print('batch times')
        for batch in padding_ds:
            t1 = time.time()
            for label in tf.unstack(model(batch)):
                label_list = tf.reshape(label, [-1]).numpy()
                labels.append(label_list)
            print(time.time()-t1)

        print('get text')
        t1 = time.time()
        if args.extract:
            for doc_id in range(len(labels)):
                text = []
                for leaf, label in zip(all_leaves[doc_id], labels[doc_id]):
                    # labels from sigmoid in [0, 1]
                    if label > 0.5:
                        text.append(leaf)
                clean_docs.extend(text)
        print(time.time()-t1)

        print('mark html')
        t1 = time.time()
        if args.save:
            for doc_id, in_file in enumerate(files):
                # path through html tree in same order as during representation building
                with open(in_file, 'rb') as hfile:
                    doc = BeautifulSoup(hfile, features='html5lib')
                    mark_leaves(doc.find_all('html')[0], all_leaves[doc_id], labels[doc_id])

                    with open("marked-" + os.path.basename(in_file), "w") as file:
                        file.write(str(doc))
        print(time.time()-t1)

    if args.extract:
        print(clean_docs)

    end = time.time()
    print('total time:')
    print(end-start)


if __name__ == '__main__':
    main()