#import os
#os.environ["CUDA_VISIBLE_DEVICES"]="-1" 
import sys
import zipfile
from collections import defaultdict
from multiprocessing import Pool

import nltk
nltk.download('punkt')
import json
import numpy as np
import tensorflow as tf
from bs4 import BeautifulSoup, Comment, NavigableString


# everything that is a child of one of these tags is considered boilerplate and hence ignored by the classifier
TAGS_TO_IGNORE = {'head', 'iframe', 'script', 'meta', 'link', 'style', 'input', 'checkbox', 'button', 'noscript'}

#copied from boilernet.net.preprocess
def get_leaves(node, tag_list=[]):
    """Return all leaves (NavigableStrings) in a BS4 tree."""
    tag_list_new = tag_list + [node.name]

    result = []
    for c in node.children:
        if isinstance(c, NavigableString):
            # might be just whitespace
            if c.string is not None and c.string.strip():
                result.append((c, tag_list_new))
        elif c.name not in TAGS_TO_IGNORE:
            result.extend(get_leaves(c, tag_list_new))
    return result

#copied from boilernet.net.preprocess
def get_leaf_representation(node, tag_list):
    """Return dicts of words and HTML tags that representat a leaf."""
    tags_dict = defaultdict(int)
    for tag in tag_list:
        tags_dict[tag] += 1
    words_dict = defaultdict(int)
    for word in nltk.word_tokenize(node.string):
        words_dict[word.lower()] += 1
    return dict(words_dict), dict(tags_dict)

def get_representation(in_file, html_archive, zip_path):
    document = []
    leaves = []
    if html_archive:
        with zipfile.ZipFile(zip_path) as z:
            with z.open(in_file, 'r') as hfile:
                doc = BeautifulSoup(hfile, features='html5lib')
                for leaf, tag_list in get_leaves(doc.find_all('html')[0]):
                    leaves.append(leaf)
                    leaf_representation = get_leaf_representation(leaf, tag_list)
                    document.append(leaf_representation)
    else:
        with open(in_file, 'rb') as hfile:
            doc = BeautifulSoup(hfile, features='html5lib')
            for leaf, tag_list in get_leaves(doc.find_all('html')[0]):
                leaves.append(leaf)
                leaf_representation = get_leaf_representation(leaf, tag_list)
                document.append(leaf_representation)

    # for pickling in multiprocessing
    leaves = [str(leaf) for leaf in leaves]
    return document, leaves

# adapted from boilernet.net.preprocess.get_leaves
def mark_leaves(node, leaves, labels):
    """Mark the leaves with score > 0.5 with the '__boilernet_label'."""
    assert len(leaves) == len(labels)

    for c in node.children:
        # redundant check
        if len(leaves) == 0 and len(labels) == 0:
            return [], []
        if isinstance(c, NavigableString):
            # might be just whitespace
            if c.string is not None and c.string.strip():
                # double check if text correct, then decide based on score
                if c == leaves[0]:
                    if labels[0] > 0.5:
                        node.attrs['__boilernet_label'] = None
                leaves.pop(0)
                labels = labels[1:]
        elif c.name not in TAGS_TO_IGNORE:
            leaves, labels = mark_leaves(c, leaves, labels)
    return leaves, labels

def main():
    num_workers = 8
    batch_size = 32

    model = tf.keras.models.load_model(sys.argv[3]) 

    with open(sys.argv[4], 'r', encoding='utf-8') as fp:
        tag_map = json.load(fp)

    with open(sys.argv[5], 'r', encoding='utf-8') as fp:
        word_map = json.load(fp)

    if sys.argv[2] == 'html':
        save = True
    elif sys.argv[2] == 'text':
        save = False
    else:
        print('Unexpected output type:', sys.argv[2])

    html_archive = False
    if sys.argv[1].endswith('.html'):
        IN_FILES = [sys.argv[1]]
    elif sys.argv[1].endswith('.zip'):
        html_archive = True
        with zipfile.ZipFile(sys.argv[1]) as z:
            IN_FILES = z.namelist()
    else:
        print('Unexpected input file type:', sys.argv[1])

    with zipfile.ZipFile('result.zip', 'w') as result:
        # only keep batch_size many inputs in memory
        for i in range(len(IN_FILES) // batch_size + 1):
            if i < len(IN_FILES) // batch_size:
                files = IN_FILES[i * batch_size : (i + 1) * batch_size]
            else:
                files = IN_FILES[i * batch_size :]

            with Pool(num_workers) as p:
                in_tuples = [(in_file, html_archive, sys.argv[1]) for in_file in files]
                all_results = p.starmap(get_representation, in_tuples)
            all_documents = [item[0] for item in all_results]
            all_leaves = [item[1] for item in all_results]

            # status bar update: first half of current step
            with open('status.txt', 'w') as stat:
                stat.write(str(round(100 * (2 * i) / (2 * (len(IN_FILES) // batch_size + 1)))))

            all_features = []
            for doc in all_documents:
                doc_features = []
                for words_dict, tags_dict in doc:
                    vocab_vec = np.zeros(len(word_map), dtype='int32')
                    for word, num in words_dict.items():
                        # if the item is not in the map, use 0 (OOV word)
                        vocab_vec[word_map.get(word, 0)] = num

                    tags_vec = np.zeros(len(tag_map), dtype='int32')
                    for tag, num in tags_dict.items():
                        # if the tag is not in the map, use 0 (OOV tag)
                        tags_vec[tag_map.get(tag, 0)] = num

                    feature_vector = np.concatenate([vocab_vec, tags_vec])
                    doc_features.append(feature_vector)

                # build tensor format for model
                doc_features = np.array(doc_features, dtype='float32', ndmin=2)
                doc_features = tf.stack(doc_features)
                all_features.append(doc_features)

            labels = []


            # create Dataset to format the batch 
            ds = tf.data.Dataset.from_generator(lambda: all_features, tf.float32, output_shapes=[None, 1052])
            padding_ds = tf.data.Dataset.padded_batch(ds, 64, [None, len(word_map.items()) + len(tag_map.items())])
            for batch in padding_ds:
                for label in tf.unstack(model(batch)):
                    label_list = tf.reshape(label, [-1]).numpy()
                    labels.append(label_list)

            # status bar update: second half of current step
            with open('status.txt', 'w') as stat:
                stat.write(str(round(100 * (2 * i + 1) / (2 * (len(IN_FILES) // batch_size + 1)))))

            if not save:
                for doc_id in range(len(labels)):
                    text = []
                    for leaf, label in zip(all_leaves[doc_id], labels[doc_id]):
                        # labels from sigmoid in [0, 1]
                        if label > 0.5:
                            text.append(leaf)

                    result.writestr(files[doc_id][:-5] + '.txt', '\n'.join(text))

            if save:
                for doc_id, in_file in enumerate(files):
                    # path through html tree in same order as during representation building
                    if html_archive:
                        with zipfile.ZipFile(sys.argv[1]) as z:
                            with z.open(in_file, 'r') as hfile:
                                doc = BeautifulSoup(hfile, features='html5lib')
                                mark_leaves(doc.find_all('html')[0], all_leaves[doc_id], labels[doc_id])

                                result.writestr(in_file, str(doc))

                    else:
                        with open(in_file, 'rb') as hfile:
                            doc = BeautifulSoup(hfile, features='html5lib')
                            mark_leaves(doc.find_all('html')[0], all_leaves[doc_id], labels[doc_id])

                            result.writestr(in_file, str(doc))

        with open('status.txt', 'w') as stat:
            stat.write(str(100))



if __name__ == '__main__':
    main()